local MAX_UNITFRAMES = 40;
local PARTY_UNITS = {"player", "party1", "party2", "party3", "party4"};
local UNITFRAME_GROUPSIZE = 5;
local UNIT_WIDTH = 125;
local UNIT_HEIGHT = 60;
local UNIT_HEALTH_HEIGHT = UNIT_HEIGHT * 35 / 60;
local UNIT_FRAMES_OFFSET_X = 51;
local UNIT_FRAMES_OFFSET_Y = -55;

HIGHLIGHT_COLOR = {r=0, g=1, b=0};
RENEWING_MIST_COLOR = {r=0, g=0.7, b=0.7};
ENVELOPING_MIST_COLOR = {r=1, g=0.7, b=0.4};
SOOTHING_MIST_COLOR = {r=0.5, g=1.0, b=0.7};

-- Stern, Kreis, ...
local raidicons = {};
raidicons[0] = {left=0, right=0, top=0, bottom=0};
raidicons[1] = {left=0, right=0.25, top=0, bottom=0.25};
raidicons[2] = {left=0.25, right=0.5, top=0, bottom=0.25};
raidicons[3] = {left=0.5, right=0.75, top=0, bottom=0.25};
raidicons[4] = {left=0.75, right=1, top=0, bottom=0.25};
raidicons[5] = {left=0, right=0.25, top=0.25, bottom=0.5};
raidicons[6] = {left=0.25, right=0.5, top=0.25, bottom=0.5};
raidicons[7] = {left=0.5, right=0.75, top=0.25, bottom=0.5};
raidicons[8] = {left=0.75, right=1, top=0.25, bottom=0.5};

-- die Farben sind etwas heller, damit sie besser erkennbar sind. Es wird ja nur das Icon eingefärbt.
local MWDebuffTypeColor = { };
MWDebuffTypeColor["none"] = { r = 0.8, g = 0, b = 0 };
MWDebuffTypeColor["Magic"] = { r = 0.4, g = 0.7, b = 1.0 };
MWDebuffTypeColor["Curse"] = { r = 0.6, g = 0.0, b = 1.0 };
MWDebuffTypeColor["Disease"] = { r = 1.0, g = 0.6, b = 0.0 };
MWDebuffTypeColor["Poison"] = { r = 0.4, g = 1.0, b = 0.4 };
MWDebuffTypeColor[""] = MWDebuffTypeColor["none"];

function MistWeaver_CreateUnitFrames()
    local unitFrame;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        if (not unitFrame) then
            unitFrame = CreateFrame("Frame", "MwUnitFrame"..i, MistWeaverFrame, "MwUnitFrameTemplate");
            MistWeaver_InitUnitFrame(unitFrame);
        end
    end

    local targetUnitFrame = CreateFrame("Frame", "MwUnitFrameTarget", MistWeaverFrame, "MwUnitFrameTemplate");
    MistWeaver_InitUnitFrame(targetUnitFrame);
    targetUnitFrame.unit = "target";

    MistWeaver_SetupUnitFramePositions();
end

function MistWeaver_UpdateUnitFrameSize(width, height)
    if (InCombatLockdown()) then
        return;
    end

    MistWeaverSettings.UNIT_WIDTH = width
    MistWeaverSettings.UNIT_HEIGHT = height

    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        if (unitFrame) then
            unitFrame:SetSize(MistWeaverSettings.UNIT_WIDTH or UNIT_WIDTH, MistWeaverSettings.UNIT_HEIGHT or UNIT_HEIGHT);
        end
    end

    local targetUnitFrame = _G["MwUnitFrameTarget"];
    
    if (targetUnitFrame) then
        targetUnitFrame:SetSize(MistWeaverSettings.UNIT_WIDTH or UNIT_WIDTH, MistWeaverSettings.UNIT_HEIGHT or UNIT_HEIGHT);
    end
end

function MistWeaver_InitUnitFrame(unitFrame)
    unitFrame:Hide();
    unitFrame:EnableMouse(false);

    local framename = unitFrame:GetName();
    unitFrame:SetFrameStrata("BACKGROUND");
    unitFrame:EnableMouse(true);
    unitFrame:SetSize(MistWeaverSettings.UNIT_WIDTH or UNIT_WIDTH, MistWeaverSettings.UNIT_HEIGHT or UNIT_HEIGHT);
    MistWeaver_SetBackdrop(unitFrame);

    -- frames

    local infoFrame = CreateFrame("Frame", framename.."InfoFrame", unitFrame, "BackdropTemplate");
    infoFrame:ClearAllPoints();
    infoFrame:SetPoint("TOPLEFT", unitFrame, "TOPLEFT", 0, 0);
    infoFrame:SetPoint("BOTTOMRIGHT", unitFrame, "BOTTOMLEFT", 30, 0);
    infoFrame:SetFrameStrata("LOW");
    infoFrame:HookScript("OnEnter", MistWeaver_OnUnitFrameInfoEnter);
    infoFrame:HookScript("OnLeave", MistWeaver_OnUnitFrameInfoLeave);
    infoFrame.unitFrame = unitFrame;
    unitFrame.infoFrame = infoFrame;

    local barFrame = CreateFrame("Frame", framename.."BarFrame", unitFrame, "BackdropTemplate");
    barFrame:ClearAllPoints();
    barFrame:SetPoint("TOPLEFT", infoFrame, "TOPRIGHT", 0, 0);
    barFrame:SetPoint("BOTTOMRIGHT", unitFrame, "BOTTOMRIGHT", 0, 0);
    barFrame:SetFrameStrata("LOW");
    unitFrame.barFrame = barFrame;

    unitFrame.hover:SetAlpha(0.1);
    unitFrame.hover:SetVertexColor(HIGHLIGHT_COLOR.r, HIGHLIGHT_COLOR.g, HIGHLIGHT_COLOR.b);

    unitFrame.aggro:SetAlpha(0.5);

    local stateFrame = CreateFrame("Frame", framename.."StateFrame", unitFrame, "MwUnitFrameStateTemplate");
    stateFrame:ClearAllPoints();
    stateFrame:SetPoint("TOPLEFT", unitFrame, "TOPLEFT", 0, 0);
    stateFrame:SetPoint("BOTTOMRIGHT", unitFrame, "BOTTOMRIGHT", 0, 0);
    stateFrame:SetFrameStrata("HIGH");
    unitFrame.stateFrame = stateFrame;    

    ------------------------------------------------------------------------
    ---- Init info frame
    ------------------------------------------------------------------------

    local classIcon = infoFrame:CreateTexture(unitFrame:GetName().."ClassIcon", "ARTWORK");
    classIcon:SetSize(22, 22);
    classIcon:ClearAllPoints();
    classIcon:SetPoint("TOP", infoFrame, "TOP", 3, -8);
    classIcon:SetTexture(nil);
    classIcon:SetAlpha(0.8);
    unitFrame.classIcon = classIcon;

    local raidIcon = infoFrame:CreateTexture(unitFrame:GetName().."RaidIcon", "ARTWORK");
    raidIcon:SetSize(22, 22);
    raidIcon:ClearAllPoints();
    raidIcon:SetPoint("TOP", infoFrame, "TOP", 3, -8);
    raidIcon:SetTexture(nil);
    raidIcon:SetAlpha(0.8);
    unitFrame.raidIcon = raidIcon;

    local detoxButtonName = unitFrame:GetName().."Detox";
    MistWeaverSpells[detoxButtonName] = 115450;
    local detoxButton = MistWeaver_CreateActionButton(detoxButtonName, id, "player", true);
    detoxButton:SetParent(infoFrame);
    detoxButton:ClearAllPoints();
    detoxButton:SetPoint("BOTTOMRIGHT", infoFrame, "BOTTOMRIGHT", -2, 3);
    detoxButton:SetScale(0.6);
    detoxButton.NormalTexture:SetTexture(nil);
    detoxButton.icon:SetTexCoord(0.05, 0.95, 0.05, 0.95);

    detoxButton.dnd = false;
    detoxButton.updateAlpha = false;
    unitFrame.detoxButton = detoxButton;

    ------------------------------------------------------------------------
    ---- Init bar frame
    ------------------------------------------------------------------------

    -- health bar

    local healthBar = CreateFrame("StatusBar", framename.."Health", barFrame, "MwTextStatusBarTemplate");
    healthBar:ClearAllPoints();
    healthBar:SetPoint("TOPLEFT", barFrame, "TOPLEFT", 3, -3);
    healthBar:SetPoint("BOTTOMRIGHT", barFrame, "BOTTOMRIGHT", -3, UNIT_HEIGHT - UNIT_HEALTH_HEIGHT);
    healthBar:SetStatusBarColor(0.0, 1.0, 0.0);
    healthBar:SetFrameStrata("LOW");

    local healthPredictionBar = CreateFrame("StatusBar", framename.."HealthPrediction", healthBar, "MwTextStatusBarTemplate");
    healthPredictionBar:ClearAllPoints();
    healthPredictionBar:SetAllPoints(healthBar);
    healthPredictionBar:SetAlpha(0.3);
    healthPredictionBar:SetStatusBarColor(0.3, 1.0, 0.7);
    healthPredictionBar:SetFrameStrata("BACKGROUND");
    healthPredictionBar:SetMinMaxValues(0, 100);
    healthPredictionBar:SetValue(0);

    local overAbsorbGlow = healthBar:CreateTexture(healthBar:GetName().."OverAbsorbGlow", "OVERLAY", "OverAbsorbGlowTemplate");
    overAbsorbGlow:SetPoint("TOPLEFT", healthBar, "TOPRIGHT", -2, 0);
    overAbsorbGlow:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", 3, 0);
    overAbsorbGlow:SetDrawLayer("OVERLAY");
    overAbsorbGlow:Hide();

    local absorbBar = healthBar:CreateTexture(healthBar:GetName().."AbsorbBar", "BACKGROUND", "TotalAbsorbBarTemplate");
    absorbBar:SetPoint("TOPLEFT", healthBar, "TOPLEFT", 0, 0);
    absorbBar:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMLEFT", 0, 0);
    absorbBar:SetDrawLayer("BACKGROUND");

    -- renewing mist bar

    local renewingMistBar = CreateFrame("StatusBar", framename.."RenewingMist", barFrame, "MwTextStatusBarTemplate");
    renewingMistBar:ClearAllPoints();
    renewingMistBar:SetPoint("TOPLEFT", healthBar, "BOTTOMLEFT", 0, 0);
    renewingMistBar:SetPoint("BOTTOMRIGHT", barFrame, "BOTTOMRIGHT", -3, 3);
    renewingMistBar:SetStatusBarColor(RENEWING_MIST_COLOR.r, RENEWING_MIST_COLOR.g, RENEWING_MIST_COLOR.b);
    renewingMistBar:SetFrameStrata("LOW");

    -- aggro bar

    local aggroBar = CreateFrame("StatusBar", framename.."Aggro", renewingMistBar, "MwTextStatusBarTemplate");
    aggroBar:ClearAllPoints();
    aggroBar:SetPoint("TOPLEFT", renewingMistBar, "BOTTOMLEFT", 3, 0);
    aggroBar:SetPoint("BOTTOMRIGHT", renewingMistBar, "BOTTOMRIGHT", 0, 0);
    aggroBar:SetStatusBarColor(1.0, 0.0, 0.0);
    aggroBar:SetFrameStrata("MEDIUM");
	aggroBar:SetStatusBarTexture("ui-castingbar-filling-standard");    

    -- soothing mist bar

    local soothingMistBar = CreateFrame("StatusBar", framename.."SoothingMist", renewingMistBar, "MwTextStatusBarTemplate");
    soothingMistBar:ClearAllPoints();
    soothingMistBar:SetPoint("TOPLEFT", healthBar, "BOTTOMLEFT", 0, 6);
    soothingMistBar:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", 0, 0);
    soothingMistBar:SetStatusBarColor(SOOTHING_MIST_COLOR.r, SOOTHING_MIST_COLOR.g, SOOTHING_MIST_COLOR.b);
    soothingMistBar:SetFrameStrata("MEDIUM");
	soothingMistBar:SetStatusBarTexture("ui-castingbar-filling-standard");

    -- enveloping mist bar

    local envelopingMistBar = CreateFrame("StatusBar", framename.."EnvelopingMist", healthBar, "MwTextStatusBarTemplate");
    envelopingMistBar:ClearAllPoints();
    envelopingMistBar:SetPoint("TOPLEFT", renewingMistBar, "TOPLEFT", 0, 0);
    envelopingMistBar:SetPoint("BOTTOMRIGHT", renewingMistBar, "TOPRIGHT", 0, -3);
    envelopingMistBar:SetStatusBarColor(ENVELOPING_MIST_COLOR.r, ENVELOPING_MIST_COLOR.g, ENVELOPING_MIST_COLOR.b);
    envelopingMistBar:SetFrameStrata("MEDIUM");
	envelopingMistBar:SetStatusBarTexture("ui-castingbar-filling-standard");

    -- Other

    local lifeCocoonIcon = healthBar:CreateTexture(framename.."LifeCocoon", "OVERLAY");
    lifeCocoonIcon:SetTexture("Interface\\Addons\\MistWeaver\\images\\lifeCocoon");
    lifeCocoonIcon:SetSize(30, 30);
    lifeCocoonIcon:ClearAllPoints();
    lifeCocoonIcon:SetPoint("CENTER", healthBar, "BOTTOM", 0, 0);
    lifeCocoonIcon:SetDrawLayer("OVERLAY");
    lifeCocoonIcon:SetAlpha(0.8);
    lifeCocoonIcon:Hide();

    -- action buttons

    local button1 = CreateFrame("Button", framename.."Button1", barFrame, "SecureActionButtonTemplate");
    button1:SetFrameStrata("LOW");
    button1:ClearAllPoints();
    button1:SetPoint("TOPLEFT", healthBar, "TOPLEFT", 0, 0);
    button1:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", 0, 0);
    button1:HookScript("OnEnter", MistWeaver_OnUnitFrameEnter);
    button1:HookScript("OnLeave", MistWeaver_OnUnitFrameLeave);
    button1:RegisterForClicks("AnyUp", "AnyDown");
    button1.unitFrame = unitFrame;

    local button2 = CreateFrame("Button", framename.."Button2", barFrame, "SecureActionButtonTemplate");
    button2:SetFrameStrata("LOW");
    button2:ClearAllPoints();
    button2:SetPoint("TOPLEFT", renewingMistBar, "TOPLEFT", 0, 0);
    button2:SetPoint("BOTTOMRIGHT", renewingMistBar, "BOTTOMRIGHT", 0, 0);
    button2:HookScript("OnEnter", MistWeaver_OnUnitFrameEnter);
    button2:HookScript("OnLeave", MistWeaver_OnUnitFrameLeave);
    button2:RegisterForClicks("AnyUp", "AnyDown");
    button2.unitFrame = unitFrame;
end

function MistWeaver_OnUnitFrameEnter(self)
    self.unitFrame.hover:Show();
end

function MistWeaver_OnUnitFrameLeave(self)
    self.unitFrame.hover:Hide();
end

function MistWeaver_OnUnitFrameInfoEnter(self)
    GameTooltip:SetUnit(self.unitFrame.unit);
end

function MistWeaver_OnUnitFrameInfoLeave(self)
    GameTooltip:FadeOut();
end

function MistWeaver_SetupUnitFramePositions()
    local unitFrame, relativeTo, realtivePoint, x, y, pos, firstInGroup;

    relativeTo = MistWeaverFrame;

    for i=1, MAX_UNITFRAMES, 1 do
        pos = mod(i, UNITFRAME_GROUPSIZE);
        unitFrame = _G["MwUnitFrame"..i];

        if (relativeTo == MistWeaverFrame) then
            realtivePoint = "TOPLEFT";
            x = UNIT_FRAMES_OFFSET_X;
            y = UNIT_FRAMES_OFFSET_Y;
        elseif (pos == 1) then
            relativeTo = firstInGroup;
            realtivePoint = "TOPRIGHT";
            x = 1;
            y = 0;
        else
            realtivePoint = "BOTTOMLEFT";
            x = 0;
            y = -2;
        end

        if (pos == 1) then
            firstInGroup = unitFrame;
        end

        unitFrame:ClearAllPoints();
        unitFrame:SetPoint("TOPLEFT", relativeTo, realtivePoint, x, y);

        relativeTo = unitFrame;
    end

    local targetUnitFrame = _G["MwUnitFrameTarget"];

    if (IsInRaid()) then
        targetUnitFrame:ClearAllPoints();
        targetUnitFrame:SetPoint("TOPLEFT", MwUnitFrame5, "BOTTOMLEFT", 0, -4);
    else
        targetUnitFrame:ClearAllPoints();
        targetUnitFrame:SetPoint("TOPLEFT", MwUnitFrame1, "TOPRIGHT", 1, 0);
    end

    if (MistWeaverSettings.TARGET_HIDDEN == false) then
        targetUnitFrame:Show();
    else
        targetUnitFrame:Hide();
    end
end

function MistWeaver_PartyMembersChanged()
    MistWeaver_RebindUnitFrames();
end

function MistWeaver_RebindUnitFrames()
    if (InCombatLockdown()) then
        return;
    end

    MistWeaver_UpdateActiveState();
    MistWeaver_SetupUnitFramePositions();

    if (IsInRaid()) then
        local units = {};
        for i=1, MAX_UNITFRAMES, 1 do
            units[i] = "raid"..i
        end

        MistWeaver_RebindUnitFramesByRole(units);
    else
        MistWeaver_RebindUnitFramesByRole(PARTY_UNITS);
    end

    MistWeaver_UnitFrames_RebindSpells(MistWeaverSpells or {});
end

function MistWeaver_ResetRebindUnitFrames()
    local unitFrame;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        MistWeaver_SetUnit(unitFrame, nil, nil);
    end
end

function MistWeaver_RebindUnitFramesByRole(units)
    MistWeaver_ResetRebindUnitFrames();

    local targetUnitFrame = _G["MwUnitFrameTarget"];
    MistWeaver_SetUnit(targetUnitFrame, "target", nil);

    local name, unitFrame, role;
    local index = 1;

    -- find tanks

    for _, id in pairs(units) do
        role = UnitGroupRolesAssigned(id);
        if (UnitExists(id) and role == "TANK") then
            unitFrame = _G["MwUnitFrame"..index];
            MistWeaver_SetUnit(unitFrame, id, role);
            index = index + 1;
        end
    end

    -- find healers

    for _, id in pairs(units) do
        role = UnitGroupRolesAssigned(id);
        if (UnitExists(id) and role == "HEALER") then
            unitFrame = _G["MwUnitFrame"..index];
            MistWeaver_SetUnit(unitFrame, id, role);
            index = index + 1;
        end
    end

    -- find damage dealers

    for _, id in pairs(units) do
        role = UnitGroupRolesAssigned(id);
        if (UnitExists(id) and role == "DAMAGER") then
            unitFrame = _G["MwUnitFrame"..index];
            MistWeaver_SetUnit(unitFrame, id, role);
            index = index + 1;
        end
    end

    -- without role

    for _, id in pairs(units) do
        role = UnitGroupRolesAssigned(id);
        if (UnitExists(id) and role == "NONE") then
            unitFrame = _G["MwUnitFrame"..index];
            MistWeaver_SetUnit(unitFrame, id, role);
            index = index + 1;
        end
    end
end

function MistWeaver_SetUnit(frame, unit, role)
    if (unit == "target" or UnitExists(unit)) then
        frame:Show();
        frame.unit = unit;
        frame.role = role;

        MistWeaver_UpdateInfoDataFor(frame, unit);
    else
        frame:Hide();
        frame.unit = nil;
        frame.role = nil;
    end

    MistWeaver_PlayerRolesAssigned();
end

------------------------------------------------------------------------
---- Spells
------------------------------------------------------------------------

function MistWeaver_UnitFrames_RebindSpells(spells)
    -- unit frames
    local unit, unitFrame, dropdown;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        unit = unitFrame.unit;

        if (unit and UnitExists(unit)) then
            MistWeaver_UnitFrames_RebindSpellForUnit("MwUnitFrame"..i, unit, spells);
        end
    end

    MistWeaver_UnitFrames_RebindSpellForUnit("MwUnitFrameTarget", "target", spells);
end

function MistWeaver_UnitFrames_RebindSpellForUnit(frameName, unit, spells)
    local button1 = _G[frameName.."Button1"];
    local button2 = _G[frameName.."Button2"];

    -- set unit
    button1:SetAttribute("unit", unit);
    button2:SetAttribute("unit", unit);

    -- no self cast and no focus cast
    button1:SetAttribute("checkselfcast", false);
    button1:SetAttribute("checkfocuscast", false);
    button2:SetAttribute("checkselfcast", false);
    button2:SetAttribute("checkfocuscast", false);

    -- spells

    button1:SetAttribute("type1", "spell");
    button1:SetAttribute("spell1", spells["mwActionHealthLeftclick"]);
    button1:SetAttribute("type2", "spell");
    button1:SetAttribute("spell2", spells["mwActionHealthRightclick"]);
    button1:SetAttribute("type3", "spell");
    button1:SetAttribute("spell3", spells["mwActionHealthMiddleclick"]);

    button1:SetAttribute("shift-type1", "spell");
    button1:SetAttribute("shift-spell1", spells["mwActionHealthShiftLeftclick"]);
    button1:SetAttribute("shift-type2", "spell");
    button1:SetAttribute("shift-spell2", spells["mwActionHealthShiftRightclick"]);
    button1:SetAttribute("shift-type3", "spell");
    button1:SetAttribute("shift-spell3", spells["mwActionHealthShiftMiddleclick"]);

    button2:SetAttribute("type1", "spell");
    button2:SetAttribute("spell1", spells["mwActionRenewingMistLeftclick"]);
    button2:SetAttribute("type2", "spell");
    button2:SetAttribute("spell2", spells["mwActionRenewingMistRightclick"]);
    button2:SetAttribute("type3", "spell");
    button2:SetAttribute("spell3", spells["mwActionRenewingMistMiddleclick"]);

    button2:SetAttribute("shift-type1", "spell");
    button2:SetAttribute("shift-spell1", spells["mwActionRenewingMistShiftLeftclick"]);
    button2:SetAttribute("shift-type2", "spell");
    button2:SetAttribute("shift-spell2", spells["mwActionRenewingMistShiftRightclick"]);
    button2:SetAttribute("shift-type3", "spell");
    button2:SetAttribute("shift-spell3", spells["mwActionRenewingMistShiftMiddleclick"]);
end

------------------------------------------------------------------------
---- Roles
------------------------------------------------------------------------

function MistWeaver_PlayerRolesAssigned()
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        MistWeaver_UpdateRole(unitFrame);
    end
end

function MistWeaver_UpdateRole(frame)
    if (not frame.unit) then
        return;
    end

    local role = UnitGroupRolesAssigned(frame.unit);
    local roleIcon = frame.roleIcon;

    if (roleIcon) then
        if (frame.role == "MAINTANK") then
            roleIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MainTankIcon");
            roleIcon:Show();
        elseif (frame.role == "MAINASSIST") then
            roleIcon:SetTexture("Interface\\GroupFrame\\UI-Group-MainAssistIcon");
            roleIcon:Show();
        elseif (role == "TANK" or role == "HEALER" or role == "DAMAGER") then
            roleIcon:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES");
            roleIcon:SetTexCoord(GetTexCoordsForRoleSmallCircle(role));
            roleIcon:Show();
        else
            roleIcon:Hide();
        end
    end
end

------------------------------------------------------------------------
---- Update
------------------------------------------------------------------------

function MistWeaver_UnitFrames_DoUpdate(self)
    -- unit frames
    local unit, unitFrame, dropdown;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        unit = unitFrame.unit;

        if (unit and UnitExists(unit)) then
            MistWeaver_UpdateUnit(unitFrame);
        end
    end

    MistWeaver_UpdateUnit(_G["MwUnitFrameTarget"]);
end

function MistWeaver_UpdateHealPrediction()
    local unitFrame, unit;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        unit = unitFrame.unit;

        if (unit and UnitExists(unit)) then
            MistWeaver_UpdateHealPredictionForUnit(unitFrame);
        end
    end
end

local MAX_INCOMING_HEAL_OVERFLOW = 1.05;
function MistWeaver_UpdateHealPredictionForUnit(unitFrame)
    if (not unitFrame) then
        return;
    end

    local unit = unitFrame.unit;

    local healthBar = _G[unitFrame:GetName().."Health"];
    local healthPredictionBar = _G[unitFrame:GetName().."HealthPrediction"];

    local allIncomingHeal = UnitGetIncomingHeals(unit) or 0;

    --Make sure we don't go too far out of the frame.
    local health = healthBar:GetValue();
    local _, maxHealth = healthBar:GetMinMaxValues();

    --See how far we're going over.
    if (health + allIncomingHeal > maxHealth * MAX_INCOMING_HEAL_OVERFLOW) then
        allIncomingHeal = maxHealth * MAX_INCOMING_HEAL_OVERFLOW - health;
    end

    healthPredictionBar:SetMinMaxValues(0, maxHealth);
    healthPredictionBar:SetValue(health + allIncomingHeal);
end

function MistWeaver_UpdateUnit(unitFrame)
    if (not unitFrame) then
        return;
    end

    local unit = unitFrame.unit;

    if (UnitExists(unit)) then
        unitFrame:SetAlpha(1);
        unitFrame.infoFrame:SetAlpha(1);
        unitFrame.barFrame:SetAlpha(1);
        MistWeaver_UpdateUnitHealth(unitFrame, unit);
        MistWeaver_UpdateUnitBuffs(unitFrame, unit);
        MistWeaver_UpdateUnitDebuffs(unitFrame, unit);
        MistWeaver_UpdateUnitThreatSituation(unitFrame, unit);
        MistWeaver_UpdateUnitState(unitFrame, unit);
    else
        unitFrame:SetAlpha(0.3);
        unitFrame.infoFrame:SetAlpha(0);
        unitFrame.barFrame:SetAlpha(0);
    end
end

------------------------------------------------------------------------
---- Health
------------------------------------------------------------------------

function MistWeaver_UpdateUnitHealth(frame, unit)
    local healthBar = _G[frame:GetName().."Health"];

    local health = UnitHealth(unit);
    local healthMax = UnitHealthMax(unit);
    local totalAbsorbs = UnitGetTotalAbsorbs(unit);

    healthBar:SetMinMaxValues(0, healthMax);
    healthBar:SetValue(health);

    local localizedClass, englishClass = UnitClass(unit);
    local classColor = RAID_CLASS_COLORS[englishClass];

    if (not classColor) then
        classColor = { r = 1.0, g = 1.0, b = 1.0, colorStr = "ffffffff" };
    end

    local unitName = GetUnitName(unit);

    if (UnitIsPVP(unit)) then
        local englishFaction, localizedFaction = UnitFactionGroup(unit);

        if (englishFaction == "Horde") then
            unitName = "|TInterface/WorldStateFrame/HordeIcon:20|t"..unitName;
        elseif (englishFaction == "Alliance") then
            unitName = "|TInterface/WorldStateFrame/AllianceIcon:20|t"..unitName;
        end
    end
    
    local RED_COLOR    = {r=1.0, g=0.0, b=0.0};
    local ORANGE_COLOR = {r=1.0, g=0.5, b=0.25};
    local YELLOW_COLOR = {r=1.0, g=1.0, b=0.0};
    local GREEN_COLOR  = {r=0.0, g=0.8, b=0.0};

    local statusBarColor = YELLOW_COLOR;
    local textColor = classColor;

    if (UnitIsFriend(unit, "player")) then
        local percent = 0;

        if (healthMax > 0) then
            percent = 100 * health / healthMax;
        end

        if (percent < 10) then
            statusBarColor = RED_COLOR;
        elseif (percent < 35) then
            statusBarColor = ORANGE_COLOR;
            local diff = (35.0 - percent) / 25.0;
            statusBarColor.g = statusBarColor.g - (diff/2.0);
            statusBarColor.b = statusBarColor.b - (diff/4.0);
        elseif (percent < 70) then
            statusBarColor = YELLOW_COLOR;
            local diff = (70.0 - percent) / 35.0;
            statusBarColor.g = statusBarColor.g - (diff/2.0);
            statusBarColor.b = statusBarColor.b + (diff/4.0);
        else
            statusBarColor = GREEN_COLOR;
            local diff = (100.0 - percent) / 30.0;
            statusBarColor.r = statusBarColor.r + diff;
            statusBarColor.g = statusBarColor.g + (diff/5.0);
        end
    elseif (UnitIsEnemy(unit, "player")) then
        statusBarColor = RED_COLOR;
    end

    healthBar:SetStatusBarColor(statusBarColor.r, statusBarColor.g, statusBarColor.b);
    local hexColor = mw_rgb2hex(textColor.r, textColor.g, textColor.b);
    healthBar.text:SetText("|cff"..hexColor..unitName.."|r");
    healthBar.text:SetDrawLayer("OVERLAY");

    -- absorbs
    local overAbsorbGlow = _G[frame:GetName().."HealthOverAbsorbGlow"];
    local absorbBar = _G[frame:GetName().."HealthAbsorbBar"];

    local maxAbsorbs = min(healthMax, totalAbsorbs + health);

    if (totalAbsorbs + health > healthMax) then
        overAbsorbGlow:Show();
    else
        overAbsorbGlow:Hide();
    end

    local healthPercent = min(1, health/healthMax);
    local left = healthBar:GetWidth() * healthPercent;

    local absorbPercent = min(1, maxAbsorbs/healthMax);
    local right = healthBar:GetWidth() * absorbPercent;

    absorbBar:SetPoint("TOPLEFT", healthBar, "TOPLEFT", left, 0);
    absorbBar:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMLEFT", right, 0);
end

------------------------------------------------------------------------
---- Buffs/Debuffs
------------------------------------------------------------------------

RENEWING_MIST_ID = 119611;
ENVELOPING_MIST_ID = 124682;
SOOTHING_MIST_ID = 115175;
LIFE_COCOON_ID = 116849;

function MistWeaver_UpdateUnitBuffs(frame, unit)
    -- renewing mist
    local renewingMistBar = _G[frame:GetName().."RenewingMist"];
    renewingMistBar:SetValue(0);

    -- enveloping mist
    local envelopingMistBar = _G[frame:GetName().."EnvelopingMist"];
    envelopingMistBar:SetValue(0);

    -- soothing mist
    local soothingMistBar = _G[frame:GetName().."SoothingMist"];
    soothingMistBar:SetValue(0);

    -- Life Cocoon
    local lifeCocoonIcon = _G[frame:GetName().."LifeCocoon"];
    local lifeCocoonIsSet = false;
    lifeCocoonIcon:Hide();

    local name, icon, count, debuffType, duration, expirationTime, unitCaster, canStealOrPurge, xy, spellId, canApplyAura;
    for ii=1, 40 do 
        name, icon, count, debuffType, duration, expirationTime, unitCaster, canStealOrPurge, xy, spellId, canApplyAura = UnitBuff(unit, ii);
        
        if (not name) then
            break;
        end

        if (name) then
            -- mw_debug(name, spellId);
        end
        
        if (spellId == RENEWING_MIST_ID) then
            local startTime = expirationTime - duration;
            renewingMistBar:SetMinMaxValues(0, expirationTime-startTime);
            renewingMistBar:SetValue(expirationTime-GetTime());

            if (unitCaster == "player") then
                renewingMistBar:SetAlpha(1);
            else
                renewingMistBar:SetAlpha(0.5);
            end
        elseif (spellId == ENVELOPING_MIST_ID) then
            local startTime = expirationTime - duration;
            envelopingMistBar:SetMinMaxValues(0, expirationTime-startTime);
            envelopingMistBar:SetValue(expirationTime-GetTime());

            if (unitCaster == "player") then
                envelopingMistBar:SetAlpha(1);
            else
                envelopingMistBar:SetAlpha(0.5);
            end
        elseif (spellId == SOOTHING_MIST_ID) then
            local startTime = expirationTime - duration;
            soothingMistBar:SetMinMaxValues(0, expirationTime-startTime);
            soothingMistBar:SetValue(expirationTime-GetTime());

            if (unitCaster == "player") then
                soothingMistBar:SetAlpha(1);
            else
                soothingMistBar:SetAlpha(0.5);
            end
        elseif (spellId == LIFE_COCOON_ID) then
            lifeCocoonIsSet = true;
        end
    end

    if (lifeCocoonIsSet) then
        lifeCocoonIcon:Show();
    else
        lifeCocoonIcon:Hide();
    end
end

function MistWeaver_UpdateUnitDebuffs(frame, unit)
    local debuffName, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId, canApplyAura, isBossDebuff, isCastByPlayer;
    for index=1, 40 do
        debuffName, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId, canApplyAura, isBossDebuff, isCastByPlayer = UnitDebuff(unit, index, "RAID");
        
        if (not debuffName) then
            frame.detoxButton:SetAlpha(0);
            break;
        end
    
        if (debuffType == "Magic" or debuffType == "Poison" or debuffType == "Disease") then
            frame.detoxButton:SetAlpha(1);
        else
            frame.detoxButton:SetAlpha(0);
        end
    end
end

------------------------------------------------------------------------
---- Aggro
------------------------------------------------------------------------

function MistWeaver_UpdateUnitThreatSituation(frame, unit)

    -- threatpct
    -- the unit's threat on the mob as a percentage of the amount required to pull aggro, scaled according to the unit's range from the mob.
    -- At 100 the unit will pull aggro. Returns 100 if the unit is tanking and nil if the unit is not on the mob's threat list.
    -- rawthreatpct
    -- the unit's threat as a percentage of the tank's current threat. Returns nil if the unit is not on the mob's threat list.


    local isTanking, status, threatpct, rawthreatpct, threatvalue = UnitDetailedThreatSituation(unit, unit.."target");

    local r, g, b = GetThreatStatusColor(status);

    if (isTanking) then
        frame.aggro:Show();
        frame.aggro:SetVertexColor(r, g, b);
    else
        frame.aggro:Hide();
    end

    if (not threatpct) then
        threatpct = 0;
    end

    if (threatpct > 100) then
        threatpct = 100;
    end

    local aggroBar = _G[frame:GetName().."Aggro"];
    aggroBar:SetValue(threatpct);
end

------------------------------------------------------------------------
---- Unit state
------------------------------------------------------------------------

function MistWeaver_UpdateUnitState(frame, unit)
    local info = frame.stateFrame.info;
    local dc = frame.stateFrame.dc;

    local isAFK = UnitIsAFK(unit);
    local isDead = UnitIsDeadOrGhost(unit);
    local isInRange = UnitInRange(unit);
    local isConnected = UnitIsConnected(unit);

    if (unit == "player" or unit == "target") then
        isInRange = 1;
    end

    if (isAFK or isDead or (not isInRange) or (not isConnected)) then
        MistWeaver_SetUnitFrameStateAlpha(frame, 0.3);
        if (not isConnected) then
            info:SetText("");
            dc:Show();
        elseif (isAFK) then
            info:SetText(AFK);
            dc:Hide();
        elseif (isDead) then
            info:SetText(DEAD);
            dc:Hide();
        elseif (not isInRange) then
            info:SetText("");
            dc:Hide();
        end
    else
        MistWeaver_SetUnitFrameStateAlpha(frame, 1.0);
        info:SetText("");
        dc:Hide();
    end
end

function MistWeaver_SetUnitFrameStateAlpha(frame, alpha)
    frame.barFrame:SetAlpha(alpha);
end

------------------------------------------------------------------------
---- Update info
------------------------------------------------------------------------

function MistWeaver_UpdateInfoData()
    local unitFrame, unit;
    for i=1, MAX_UNITFRAMES, 1 do
        unitFrame = _G["MwUnitFrame"..i];
        unit = unitFrame.unit;

        if (unit and UnitExists(unit)) then
            MistWeaver_UpdateInfoDataFor(unitFrame, unit);
        end
    end

    MistWeaver_UpdateInfoDataFor(_G["MwUnitFrameTarget"], "target");
end

function MistWeaver_UpdateInfoDataFor(frame, unit)
    MistWeaver_UpdateClassIcon(frame, unit);
    MistWeaver_UpdateRaidIcon(frame, unit);
end

function MistWeaver_UpdateClassIcon(frame, unit)
    local icon = _G[frame:GetName().."ClassIcon"];

    if (icon) then
        local _, classToken = UnitClass(unit);

        if (classToken) then
            local coords = CLASS_ICON_TCOORDS[classToken];
            icon:SetTexture("Interface\\WorldStateFrame\\Icons-Classes");
            icon:SetTexCoord(coords[1], coords[2], coords[3], coords[4]);
        else
            icon:SetTexture(nil);
        end
    end
end

function MistWeaver_UpdateRaidIcon(frame, unit)
    local icon = _G[frame:GetName().."RaidIcon"];
    local index = GetRaidTargetIndex(frame.unit);

    local classIcon = _G[frame:GetName().."ClassIcon"];

    if (icon and classIcon) then
        if (index and index > 0) then
            if (index > 8) then
                index = 0;
            end

            icon:SetTexture("Interface\\targetingframe\\ui-raidtargetingicons");
            icon:SetTexCoord(raidicons[index].left, raidicons[index].right, raidicons[index].top, raidicons[index].bottom);
            icon:Show();
            classIcon:Hide();
        else
            icon:Hide();
            classIcon:Show();
        end
    end
end
